#include "main.hpp"
#include "secrets.hpp"

#define WIRE Wire

#define SERVO_PIN 14     // D5
#define BUZZER_PIN 12    // D6
#define LEDSTRIPE_PIN 15 // D8

// buttons
#define TACTILE_BTN_LEFT 0  // D3
#define TACTILE_BTN_RIGHT 2 // D4
#define TOUCH_BTN_PIN 16    // D0

// RFID
#define PN532_IRQ 13   // D7 (not connected)
#define PN532_RESET 12 // D6 (not connected)
// Or use hardware Serial:
Adafruit_PN532 nfc(PN532_IRQ, PN532_RESET); // RFID with I2C

// OLED display
#define DISPLAY_PIN 4 // D2 (SDA)
auto display = Adafruit_SSD1306(128, 32, &WIRE);

// util classes
SmartStrip smartStrip;
SmartServo smartServo(SERVO_PIN);
TeaTimer teaTimer(&smartServo);
SmartRFID smartRFID(&nfc);
SmartDisplay smartDisplay(&display);
TeaNetworking teaNet;

SmartSound sound(BUZZER_PIN);
SmartFeedback smartFeedback;

SmartButton buttonLeft(TACTILE_BTN_LEFT);
SmartButton buttonRight(TACTILE_BTN_RIGHT);

// networking
String shuttleServer = "https://ias-tea-axum.shuttleapp.rs";

TeaData currentTea(-1, "Teafault", "c0derfid", "", -1, 30);

// config
long feedbackTimeout = 15;
int defaultSteepingTime = 120;

// timeout
unsigned long timeoutTime;

// macros
#define CATCH_NET_FAILURE(netResult, msg) \
  if (netResult == JSON.parse("{}"))      \
  {                                       \
    smartDisplay.printMsg(msg);           \
    delay(500);                           \
    break;                                \
  }

enum State
{
  // lifecycle states
  WAIT_FOR_RFID_SCAN,
  PROCESS_RFID_SCAN,
  DISPLAY_SCANNED_TEA_CONFIG,
  STEEPING_IN_PROGRESS,
  REQUEST_FEEDBACK,

  // special states
  DISPLAY_ADD_NEW_TEA_CONFIG,

  // debugging states
  NETWORKING_DEBUG,
  SERVO_TUNING,
};
State state = State::WAIT_FOR_RFID_SCAN;

void setup()
{
  Serial.begin(9600);

  smartServo.moveServoToZero();

  smartStrip.init();

  smartStrip.setGreen(0);
  smartStrip.setGreen(7);
  smartStrip.setGreen(14);

  smartRFID.init();

  Serial.println("Initializing OLED...");
  display.begin(SSD1306_SWITCHCAPVCC, 0x3C, false); // Address 0x3C for 128x64

  smartDisplay.playTeaAnimationPartOne(32, "Waiting", "for", "internet", "connection...");

  teaNet.init(ssid, password);

  smartDisplay.playTeaAnimationPartTwo(32, "", "Success!", "", "");

  // load config
  JSONVar result = teaNet.httpsGETRequest(shuttleServer + "/configuration");
  // catch possible failure
  if (result == JSON.parse("{}"))
  {
    smartDisplay.printMsg("failed loading config");
    delay(500);
  }
  else
  {
    defaultSteepingTime = result["default_steeping_time"];
    feedbackTimeout = result["feedback_timeout"];
  }

  smartDisplay.playTeaAnimation();
  delay(250);

  pinMode(LED_BUILTIN, OUTPUT);
  pinMode(TOUCH_BTN_PIN, INPUT);

  buttonLeft.init();
  buttonRight.init();
}

u_int frame = 0;
u_int flicker = 0;
void loop()
{

  switch (state)
  {
  case SERVO_TUNING:
  {
    // default min 1000, max 2000
    smartDisplay.printMsg("Write servo 0");
    smartServo.myservo.write(0);
    delay(2000);

    smartDisplay.printMsg("Write servo 180");
    smartServo.myservo.write(180);
    delay(2000);
    break;
  }
  case NETWORKING_DEBUG:
  {
    if (!teaNet.wifiConnected())
    {
      state = WAIT_FOR_RFID_SCAN;
      break;
    }
    JSONVar jsonData = teaNet.httpsGETRequest(shuttleServer + "/types-of-tea");
    Serial.println("JSON:");
    Serial.println(jsonData[0]["tea_name"]);
    delay(2000);

    String payload = teaNet.bundleNewTeaInfo("espDemoRfid", 180);
    teaNet.httpsPOSTRequest("https://ias-tea-axum.shuttleapp.rs/add-tea", payload);

    state = WAIT_FOR_RFID_SCAN;
    break;
  }
  case WAIT_FOR_RFID_SCAN:
  {
    smartDisplay.playRfidWaitAnimation(frame++);

    smartStrip.setGreen(0);
    smartStrip.setGreen(7);
    smartStrip.setGreen(14);

    // this blocks execution for 100ms to check for a tag
    String uid = smartRFID.readTag(100);
    if (uid != "no_tag_detected" && uid != "")
    {
      currentTea.m_rfidCode = uid;
      state = PROCESS_RFID_SCAN;
    }
    break;
  }
  case PROCESS_RFID_SCAN:
  {
    smartDisplay.playTeaAnimationPartOne(32, "", "Loading...", "", "");

    JSONVar result = teaNet.httpsGETRequest(shuttleServer + "/tea-by-rfid/" + currentTea.m_rfidCode);

    CATCH_NET_FAILURE(result, "GET FAILED")

    if (JSONVar::stringify(result) == "null")
    {
      Serial.print("TEA DOES NOT EXIST YET!");
      state = State::DISPLAY_ADD_NEW_TEA_CONFIG;
      currentTea = TeaData(-1, "New Tea", currentTea.m_rfidCode, "", -1, defaultSteepingTime);
      smartDisplay.playTeaAnimationPartTwo(32, "", "Done!", "", "");
      break;
    }
    else
    {
      state = State::DISPLAY_SCANNED_TEA_CONFIG;
      currentTea = TeaData(result["id"], result["tea_name"], result["rfid_code"], result["tea_notes"], result["water_temp"], result["steeping_seconds"]);
      smartDisplay.playTeaAnimationPartTwo(32, "", "Done!", "", "");
    }

    break;
  }
  case DISPLAY_SCANNED_TEA_CONFIG:
  {

    String uid = smartRFID.readTag(50); // just enough time to detect tag
    if (uid != "no_tag_detected" && uid != "")
    {
      currentTea.m_rfidCode = uid;
      smartDisplay.resetDisplay();
      state = PROCESS_RFID_SCAN;
      break;
    }

    if (buttonLeft.isJustPressed())
    {
      currentTea.m_steepingSeconds -= 10;
      if (currentTea.m_steepingSeconds < 0)
      {
        currentTea.m_steepingSeconds = 0;
      }
    }
    if (buttonRight.isJustPressed())
    {
      currentTea.m_steepingSeconds += 10;
    }
    smartDisplay.printTeaConfigScreen(currentTea);

    if (isTouchBtnPressed() && currentTea.m_steepingSeconds <= 0)
    {
      // rickroll
      Serial.println("Easter");
      sound.play(SmartSound::NeverGonna, 155);
      break;
    }

    // change state if start pressed
    if (isTouchBtnPressed())
    {
      teaTimer.beginSteeping(currentTea.m_steepingSeconds * 1000);
      state = State::STEEPING_IN_PROGRESS;
      smartStrip.reset();
    }
    break;
  }
  case STEEPING_IN_PROGRESS:
  {
    teaTimer.tick();
    if (teaTimer.isSteeping())
    {
      // abort if red button pressed
      if (buttonLeft.isJustPressed())
      {
        teaTimer.endSteeping();
        smartDisplay.printMsg("Ready!");
        smartStrip.animateBottomToTop(0, 1, 0);
        delay(1000);
        state = State::REQUEST_FEEDBACK;
        timeoutTime = millis() + (feedbackTimeout * 1000 * 60);
        break;
      }

      auto r = teaTimer.remainingSteepTimeSeconds();
      smartDisplay.printTeaSteepingProgressScreen(currentTea, r);

      // update led strip
      smartStrip.progressBar(0, currentTea.m_steepingSeconds, r);
    }
    else
    {
      smartDisplay.printMsg("Ready!");
      smartStrip.animateBottomToTop(0, 1, 0);
      state = State::REQUEST_FEEDBACK;
      timeoutTime = millis() + (feedbackTimeout * 1000 * 60);
    }
    break;
  }
  case REQUEST_FEEDBACK:
  {

    // timeout
    if (millis() >= timeoutTime)
    {
      // we still want to save the last configuration (treated as perfect, no steeping time change)
      String steeepingTimePayload = teaNet.bundleSteepingTimeLogEntry(currentTea.m_steepingSeconds);
      String teaMetaPayload = teaNet.bundleTeaMetaChange(currentTea.m_teaName, currentTea.m_teaNotes, currentTea.m_waterTemp, currentTea.m_steepingSeconds);
      Serial.println(steeepingTimePayload);
      smartDisplay.playTeaAnimationPartOne(32, "", "Loading...", "", "");
      JSONVar res = teaNet.httpsPOSTRequest(shuttleServer + "/tea-steeping-time-log/" + currentTea.m_id, steeepingTimePayload);
      JSONVar res2 = teaNet.httpsPUTRequest(shuttleServer + "/types-of-tea/" + currentTea.m_id, teaMetaPayload);

      CATCH_NET_FAILURE(res, "TRY AGAIN")
      CATCH_NET_FAILURE(res2, "TRY AGAIN")

      state = State::WAIT_FOR_RFID_SCAN;
      smartStrip.reset();
      break;
    }

    smartDisplay.printRequestFeedbackScreen(currentTea);

    String uid = smartRFID.readTag(50); // just enough time to detect tag
    if (uid != "no_tag_detected" && uid != "")
    {

      // we still want to save the last configuration (treated as perfect, no steeping time change)
      String steeepingTimePayload = teaNet.bundleSteepingTimeLogEntry(currentTea.m_steepingSeconds);
      String teaMetaPayload = teaNet.bundleTeaMetaChange(currentTea.m_teaName, currentTea.m_teaNotes, currentTea.m_waterTemp, currentTea.m_steepingSeconds);
      Serial.println(steeepingTimePayload);
      smartDisplay.playTeaAnimationPartOne(32, "", "Loading...", "", "");
      JSONVar res = teaNet.httpsPOSTRequest(shuttleServer + "/tea-steeping-time-log/" + currentTea.m_id, steeepingTimePayload);
      JSONVar res2 = teaNet.httpsPUTRequest(shuttleServer + "/types-of-tea/" + currentTea.m_id, teaMetaPayload);

      CATCH_NET_FAILURE(res, "TRY AGAIN")
      CATCH_NET_FAILURE(res2, "TRY AGAIN")

      currentTea.m_rfidCode = uid;
      smartDisplay.resetDisplay();
      smartStrip.reset();
      state = PROCESS_RFID_SCAN;
      break;
    }

    if (isTouchBtnPressed())
    {
      smartDisplay.playTeaAnimationPartOne(32, "Accepted!", "(perfect)", "", "saving...");

      // keep steeping time for next brew
      String steeepingTimePayload = teaNet.bundleSteepingTimeLogEntry(currentTea.m_steepingSeconds);
      String teaMetaPayload = teaNet.bundleTeaMetaChange(currentTea.m_teaName, currentTea.m_teaNotes, currentTea.m_waterTemp, currentTea.m_steepingSeconds);
      Serial.println(steeepingTimePayload);
      JSONVar res = teaNet.httpsPOSTRequest(shuttleServer + "/tea-steeping-time-log/" + currentTea.m_id, steeepingTimePayload);
      JSONVar res2 = teaNet.httpsPUTRequest(shuttleServer + "/types-of-tea/" + currentTea.m_id, teaMetaPayload);

      CATCH_NET_FAILURE(res, "TRY AGAIN")
      CATCH_NET_FAILURE(res2, "TRY AGAIN")

      state = State::WAIT_FOR_RFID_SCAN;
      smartDisplay.playTeaAnimationPartTwo(32, "", "Done!", "", "");
      smartStrip.reset();
    }
    if (buttonLeft.isJustPressed())
    {
      // decrease steeping time for next brew
      smartDisplay.playTeaAnimationPartOne(32, "Accepted!", "(shorter)", "", "saving...");

      // request steeping log
      JSONVar steepingLog = teaNet.httpsGETRequest(shuttleServer + "/brewing-events/" + currentTea.m_id);
      CATCH_NET_FAILURE(steepingLog, "TRY AGAIN")
      

      String steeepingTimePayload = teaNet.bundleSteepingTimeLogEntry(currentTea.m_steepingSeconds);
      Serial.println(steeepingTimePayload);
      // log old steeping time
      JSONVar res = teaNet.httpsPOSTRequest(shuttleServer + "/tea-steeping-time-log/" + currentTea.m_id, steeepingTimePayload);

      // now update steeping time for next use
      currentTea.m_steepingSeconds = smartFeedback.calculateNewTime(false, steepingLog, currentTea.m_steepingSeconds);
      String teaMetaPayload = teaNet.bundleTeaMetaChange(currentTea.m_teaName, currentTea.m_teaNotes, currentTea.m_waterTemp, currentTea.m_steepingSeconds);
      JSONVar res2 = teaNet.httpsPUTRequest(shuttleServer + "/types-of-tea/" + currentTea.m_id, teaMetaPayload);

      CATCH_NET_FAILURE(res, "TRY AGAIN")
      CATCH_NET_FAILURE(res2, "TRY AGAIN")

      state = State::WAIT_FOR_RFID_SCAN;
      smartDisplay.playTeaAnimationPartTwo(32, "", "Done!", "", "");
      smartStrip.reset();
    }
    if (buttonRight.isJustPressed())
    {
      // increase steeping time for next brew
      smartDisplay.playTeaAnimationPartOne(32, "Accepted!", "(longer)", "", "saving...");

      // request steeping log
      JSONVar steepingLog = teaNet.httpsGETRequest(shuttleServer + "/brewing-events/" + currentTea.m_id);
      CATCH_NET_FAILURE(steepingLog, "TRY AGAIN")

      String steeepingTimePayload = teaNet.bundleSteepingTimeLogEntry(currentTea.m_steepingSeconds);
      Serial.println(steeepingTimePayload);
      JSONVar res = teaNet.httpsPOSTRequest(shuttleServer + "/tea-steeping-time-log/" + currentTea.m_id, steeepingTimePayload);

      // now update steeping time for next use
      currentTea.m_steepingSeconds = smartFeedback.calculateNewTime(true, steepingLog, currentTea.m_steepingSeconds);
      String teaMetaPayload = teaNet.bundleTeaMetaChange(currentTea.m_teaName, currentTea.m_teaNotes, currentTea.m_waterTemp, currentTea.m_steepingSeconds);
      JSONVar res2 = teaNet.httpsPUTRequest(shuttleServer + "/types-of-tea/" + currentTea.m_id, teaMetaPayload);

      CATCH_NET_FAILURE(res, "TRY AGAIN")
      CATCH_NET_FAILURE(res2, "TRY AGAIN")

      state = State::WAIT_FOR_RFID_SCAN;
      smartDisplay.playTeaAnimationPartTwo(32, "", "Done!", "", "");
      smartStrip.reset();
    }
    break;
  }
  case DISPLAY_ADD_NEW_TEA_CONFIG:
  {
    flicker++;
    smartDisplay.printAddNewTeaConfigScreen(currentTea, &flicker);

    String uid = smartRFID.readTag(50); // just enough time to detect tag
    if (uid != "no_tag_detected" && uid != "")
    {
      currentTea.m_rfidCode = uid;
      smartDisplay.resetDisplay();
      state = PROCESS_RFID_SCAN;
      break;
    }

    if (buttonLeft.isJustPressed())
    {
      currentTea.m_steepingSeconds -= 10;
      if (currentTea.m_steepingSeconds < 0)
      {
        currentTea.m_steepingSeconds = 0;
      }
    }
    if (buttonRight.isJustPressed())
    {
      currentTea.m_steepingSeconds += 10;
    }

    if (isTouchBtnPressed())
    {
      smartDisplay.playTeaAnimationPartOne(32, "Accepted!", "", "", "saving...");

      String payload = teaNet.bundleNewTeaInfo(currentTea.m_rfidCode, currentTea.m_steepingSeconds);
      JSONVar result = teaNet.httpsPOSTRequest(shuttleServer + "/add-tea", payload);

      CATCH_NET_FAILURE(result, "POST FAILED")

      smartDisplay.playTeaAnimationPartTwo(32, "", "Done!", "", "");
      state = State::DISPLAY_SCANNED_TEA_CONFIG;
    }

    break;
  }
  };
}

bool isTouchBtnPressed()
{
  return digitalRead(TOUCH_BTN_PIN) == HIGH;
}